@@ -35,6 +35,7 @@ int Assemble(char *mddev, int mdfd,
 	     mddev_ident_t ident, char *conffile,
 	     mddev_dev_t devlist,
 	     int readonly, int runstop,
+	     char *update,
 	     int verbose, int force)
 {
 	/*
@@ -72,7 +73,7 @@ int Assemble(char *mddev, int mdfd,
 	 * For each device:
 	 *   Check superblock - discard if bad
 	 *   Check uuid (set if we don't have one) - discard if no match
-	 *   Check superblock similarity if we have a superbloc - discard if different
+	 *   Check superblock similarity if we have a superblock - discard if different
 	 *   Record events, devicenum, utime
 	 * This should give us a list of devices for the array
 	 * We should collect the most recent event and utime numbers
@@ -102,16 +103,21 @@ int Assemble(char *mddev, int mdfd,
 		long long events;
 		time_t utime;
 		int uptodate;
+		int state;
 		int raid_disk;
-	} devices[MD_SB_DISKS];
-	int best[MD_SB_DISKS]; /* indexed by raid_disk */
+	} *devices;
+	int *best = NULL; /* indexed by raid_disk */
+	int bestcnt = 0;
 	int devcnt = 0, okcnt, sparecnt;
+	int req_cnt;
 	int i;
 	int most_recent = 0;
 	int chosen_drive;
 	int change = 0;
 	int inargv = 0;
 	int start_partial_ok = force || devlist==NULL;
+	int num_devs;
+	mddev_dev_t tmpdev;
 	
 	vers = md_get_version(mdfd);
 	if (vers <= 0) {
@@ -153,8 +159,16 @@ int Assemble(char *mddev, int mdfd,
 		devlist = conf_get_devs(conffile);
 	else inargv = 1;
 
+	tmpdev = devlist; num_devs = 0;
+	while (tmpdev) {
+		num_devs++;
+		tmpdev = tmpdev->next;
+	}
+	best = malloc(num_devs * sizeof(*best));
+	devices = malloc(num_devs * sizeof(*devices));
+
 	first_super.md_magic = 0;
-	for (i=0; i<MD_SB_DISKS; i++)
+	for (i=0; i<num_devs; i++)
 		best[i] = -1;
 
 	if (verbose)
@@ -172,8 +186,11 @@ int Assemble(char *mddev, int mdfd,
 		devlist = devlist->next;
 
 		if (ident->devices &&
-		    !match_oneof(ident->devices, devname))
+		    !match_oneof(ident->devices, devname)) {
+			if (inargv || verbose)
+				fprintf(stderr, Name ": %s is not one of %s\n", devname, ident->devices);
 			continue;
+		}
 		
 		dfd = open(devname, O_RDONLY, 0);
 		if (dfd < 0) {
@@ -245,11 +262,48 @@ int Assemble(char *mddev, int mdfd,
 			return 1;
 		}
 
+
+		/* this is needed until we get a more relaxed super block format */
 		if (devcnt >= MD_SB_DISKS) {
 		    fprintf(stderr, Name ": ouch - too many devices appear to be in this array. Ignoring %s\n",
 			    devname);
 		    continue;
 		}
+		
+		/* looks like a good enough match to update the super block if needed */
+		if (update) {
+			if (strcmp(update, "sparc2.2")==0 ) {
+				/* 2.2 sparc put the events in the wrong place
+				 * So we copy the tail of the superblock
+				 * up 4 bytes before continuing
+				 */
+				__u32 *sb32 = (__u32*)&super;
+				memcpy(sb32+MD_SB_GENERIC_CONSTANT_WORDS+7,
+				       sb32+MD_SB_GENERIC_CONSTANT_WORDS+7+1,
+				       (MD_SB_WORDS - (MD_SB_GENERIC_CONSTANT_WORDS+7+1))*4);
+				fprintf (stderr, Name ": adjusting superblock of %s for 2.2/sparc compatability.\n",
+					 devname);
+			}
+			if (strcmp(update, "super-minor") ==0) {
+				struct stat stb2;
+				fstat(mdfd, &stb2);
+				super.md_minor = MINOR(stb2.st_rdev);
+				if (verbose)
+					fprintf(stderr, Name ": updating superblock of %s with minor number %d\n",
+						devname, super.md_minor);
+			}
+			super.sb_csum = calc_sb_csum(&super);
+			dfd = open(devname, O_RDWR, 0);
+			if (dfd < 0) 
+				fprintf(stderr, Name ": Cannot open %s for superblock update\n",
+					devname);
+			else if (store_super(dfd, &super))
+				fprintf(stderr, Name ": Could not re-write superblock on %s.\n",
+					devname);
+			if (dfd >= 0)
+				close(dfd);
+		}
+
 		if (verbose)
 			fprintf(stderr, Name ": %s is identified as a member of %s, slot %d.\n",
 				devname, mddev, super.this_disk.raid_disk);
@@ -262,17 +316,35 @@ int Assemble(char *mddev, int mdfd,
 		devices[devcnt].utime = super.utime;
 		devices[devcnt].raid_disk = super.this_disk.raid_disk;
 		devices[devcnt].uptodate = 0;
+		devices[devcnt].state = super.this_disk.state;
 		if (most_recent < devcnt) {
 			if (devices[devcnt].events
 			    > devices[most_recent].events)
 				most_recent = devcnt;
 		}
-		i = devices[devcnt].raid_disk;
-		if (i>=0 && i < MD_SB_DISKS)
+		if (super.level == -4) 
+			/* with multipath, the raid_disk from the superblock is meaningless */
+			i = devcnt;
+		else
+			i = devices[devcnt].raid_disk;
+		if (i>=0 && i < 10000) {
+			if (i >= bestcnt) {
+				int newbestcnt = i+10;
+				int *newbest = malloc(sizeof(int)*newbestcnt);
+				int c;
+				for (c=0; c < newbestcnt; c++)
+					if (c < bestcnt)
+						newbest[c] = best[c];
+					else
+						newbest[c] = -1;
+				if (best)free(best);
+				best = newbest;
+				bestcnt = newbestcnt;
+			}
 			if (best[i] == -1
 			    || devices[best[i]].events < devices[devcnt].events)
 				best[i] = devcnt;
-
+		}
 		devcnt++;
 	}
 
@@ -286,10 +358,19 @@ int Assemble(char *mddev, int mdfd,
 	 */
 	okcnt = 0;
 	sparecnt=0;
-	for (i=0; i< MD_SB_DISKS;i++) {
+	for (i=0; i< bestcnt ;i++) {
 		int j = best[i];
 		int event_margin = !force;
 		if (j < 0) continue;
+		/* note: we ignore error flags in multipath arrays
+		 * as they don't make sense
+		 */
+		if (first_super.level != -4)
+			if (!(devices[j].state & (1<<MD_DISK_SYNC))) {
+				if (!(devices[j].state & (1<<MD_DISK_FAULTY)))
+					sparecnt++;
+				continue;
+			}
 		if (devices[j].events+event_margin >=
 		    devices[most_recent].events) {
 			devices[j].uptodate = 1;
@@ -306,7 +387,7 @@ int Assemble(char *mddev, int mdfd,
 		 */
 		int fd;
 		chosen_drive = -1;
-		for (i=0; i<first_super.raid_disks; i++) {
+		for (i=0; i<first_super.raid_disks && i < bestcnt; i++) {
 			int j = best[i];
 			if (j>=0 &&
 			    !devices[j].uptodate &&
@@ -337,6 +418,10 @@ int Assemble(char *mddev, int mdfd,
 		}
 		super.events_hi = (devices[most_recent].events>>32)&0xFFFFFFFF;
 		super.events_lo = (devices[most_recent].events)&0xFFFFFFFF;
+		if (super.level == 5 || super.level == 4) {
+			/* need to force clean */
+			super.state = (1<<MD_SB_CLEAN);
+		}
 		super.sb_csum = calc_sb_csum(&super);
 /*DRYRUN*/	if (store_super(fd, &super)) {
 			close(fd);
@@ -358,7 +443,7 @@ int Assemble(char *mddev, int mdfd,
 	 * superblock.
 	 */
 	chosen_drive = -1;
-	for (i=0; chosen_drive < 0 && i<MD_SB_DISKS; i++) {
+	for (i=0; chosen_drive < 0 && i<bestcnt; i++) {
 		int j = best[i];
 		int fd;
 		if (j<0)
@@ -380,7 +465,7 @@ int Assemble(char *mddev, int mdfd,
 		close(fd);
 	}
 
-	for (i=0; i<MD_SB_DISKS; i++) {
+	for (i=0; i<bestcnt; i++) {
 		int j = best[i];
 		int desired_state;
 
@@ -428,6 +513,11 @@ This doesnt work yet
 				i, mddev);
 		}
 	}
+	if (force && (super.level == 4 || super.level == 5) && 
+	    okcnt == super.raid_disks-1) {
+		super.state = (1<< MD_SB_CLEAN);
+		change |= 2;
+	}
 
 	if ((force && (change & 2))
 	    || (old_linux && (change & 1))) {
@@ -449,6 +539,17 @@ This doesnt work yet
 		change = 0;
 	}
 
+	/* count number of in-sync devices according to the superblock.
+	 * We must have this number to start the array without -s or -R
+	 */
+	req_cnt = 0;
+	for (i=0; i<MD_SB_DISKS; i++)
+		if ((first_super.disks[i].state & (1<<MD_DISK_SYNC)) &&
+		    (first_super.disks[i].state & (1<<MD_DISK_ACTIVE)) &&
+		    !(first_super.disks[i].state & (1<<MD_DISK_FAULTY)))
+			req_cnt ++;
+									    
+
 	/* Almost ready to actually *do* something */
 	if (!old_linux) {
 		if (ioctl(mdfd, SET_ARRAY_INFO, NULL) != 0) {
@@ -457,16 +558,16 @@ This doesnt work yet
 			return 1;
 		}
 		/* First, add the raid disks, but add the chosen one last */
-		for (i=0; i<=MD_SB_DISKS; i++) {
+		for (i=0; i<= bestcnt; i++) {
 			int j;
-			if (i < MD_SB_DISKS) {
+			if (i < bestcnt) {
 				j = best[i];
 				if (j == chosen_drive)
 					continue;
 			} else
 				j = chosen_drive;
 
-			if (j >= 0 && devices[j].uptodate) {
+			if (j >= 0 /* && devices[j].uptodate */) {
 				mdu_disk_info_t disk;
 				memset(&disk, 0, sizeof(disk));
 				disk.major = devices[j].major;
@@ -490,12 +591,14 @@ This doesnt work yet
 		
 		if (runstop == 1 ||
 		    (runstop == 0 && 
-		     ( first_super.raid_disks == okcnt
-		       || (start_partial_ok && enough(first_super.level, first_super.raid_disks, okcnt)))
-			    )) {
+		     ( enough(first_super.level, first_super.raid_disks, okcnt) &&
+		       (okcnt >= req_cnt || start_partial_ok)
+			     ))) {
 			if (ioctl(mdfd, RUN_ARRAY, NULL)==0) {
 				fprintf(stderr, Name ": %s has been started with %d drive%s",
 					mddev, okcnt, okcnt==1?"":"s");
+				if (okcnt < first_super.raid_disks) 
+					fprintf(stderr, " (out of %d)", first_super.raid_disks);
 				if (sparecnt)
 					fprintf(stderr, " and %d spare%s", sparecnt, sparecnt==1?"":"s");
 				fprintf(stderr, ".\n");
@@ -510,8 +613,18 @@ This doesnt work yet
 				mddev, okcnt, okcnt==1?"":"s");
 			return 0;
 		}
-		fprintf(stderr, Name ": %s assembled from %d drive%s - not enough to start it (use --run to insist).\n",
-			mddev, okcnt, okcnt==1?"":"s");
+		fprintf(stderr, Name ": %s assembled from %d drive%s", mddev, okcnt, okcnt==1?"":"s");
+		if (sparecnt)
+			fprintf(stderr, " and %d spare%s", sparecnt, sparecnt==1?"":"s");
+		if (!enough(first_super.level, first_super.raid_disks, okcnt))
+			fprintf(stderr, " - not enough to start the array.\n");
+		else {
+			if (req_cnt == first_super.raid_disks)
+				fprintf(stderr, " - need all %d to start it", req_cnt);
+			else
+				fprintf(stderr, " - need %d of %d to start", req_cnt, first_super.raid_disks);
+			fprintf(stderr, " (use --run to insist).\n");
+		}
 		return 1;
 	} else {
 		/* The "chosen_drive" is a good choice, and if necessary, the superblock has
