@@ -45,7 +45,7 @@ static char *percentalerts[] = {
 
 int Monitor(mddev_dev_t devlist,
 	    char *mailaddr, char *alert_cmd,
-	    int period, int scan,
+	    int period, int daemonise, int scan, int oneshot,
 	    char *config)
 {
 	/*
@@ -105,12 +105,39 @@ int Monitor(mddev_dev_t devlist,
 	int finished = 0;
 	struct mdstat_ent *mdstat = NULL;
 
-	if (!mailaddr && scan)
+	if (!mailaddr) {
 		mailaddr = conf_get_mailaddr(config);
-	if (!alert_cmd && scan)
+		if (mailaddr && ! scan)
+			fprintf(stderr, Name ": Monitor using email address \"%s\" from config file\n",
+			       mailaddr);
+	}
+	if (!alert_cmd) {
 		alert_cmd = conf_get_program(config);
-	if (scan && !mailaddr && !alert_cmd)
-		return 0;
+		if (alert_cmd && ! scan)
+			fprintf(stderr, Name ": Monitor using program \"%s\" from config file\n",
+			       alert_cmd);
+	}
+	if (scan && !mailaddr && !alert_cmd) {
+		fprintf(stderr, Name ": No mail address or alert command - not monitoring.\n");
+		return 1;
+	}
+
+	if (daemonise) {
+		int pid = fork();
+		if (pid > 0) {
+			printf("%d\n", pid);
+			return 0;
+		}
+		if (pid < 0) {
+			perror("daemonise");
+			return 1;
+		}
+		close(0);
+		open("/dev/null", 3);
+		dup2(0,1);
+		dup2(0,2);
+		setsid();
+	}
 
 	if (devlist == NULL) {
 		mddev_ident_t mdlist = conf_get_ident(config, NULL);
@@ -121,7 +148,7 @@ int Monitor(mddev_dev_t devlist,
 			st->devname = strdup(mdlist->devname);
 			st->utime = 0;
 			st->next = statelist;
-			st->err = 1;
+			st->err = 0;
 			st->devnum = -1;
 			st->percent = -2;
 			if (mdlist->spare_group)
@@ -139,7 +166,7 @@ int Monitor(mddev_dev_t devlist,
 			st->devname = strdup(dv->devname);
 			st->utime = 0;
 			st->next = statelist;
-			st->err = 1;
+			st->err = 0;
 			st->devnum = -1;
 			st->percent = -2;
 			st->spare_group = NULL;
@@ -149,6 +176,7 @@ int Monitor(mddev_dev_t devlist,
 
 
 	while (! finished) {
+		int new_found = 0;
 		struct state *st;
 
 		if (mdstat)
@@ -214,6 +242,12 @@ int Monitor(mddev_dev_t devlist,
 				st->err = 0;
 				continue;
 			}
+			if (st->utime == 0 && /* new array */
+			    mse &&	/* is in /proc/mdstat */
+			    mse->pattern && strchr(mse->pattern, '_') /* degraded */
+				)
+				alert("DegradedArray", dev, NULL, mailaddr, alert_cmd);
+
 			if (mse &&
 			    st->percent == -1 && 
 			    mse->percent >= 0)
@@ -296,6 +330,7 @@ int Monitor(mddev_dev_t devlist,
 					st->spare_group = NULL;
 					statelist = st;
 					alert("NewArray", st->devname, NULL, mailaddr, alert_cmd);
+					new_found = 1;
 				}
 		}
 		/* If an array has active < raid && spare == 0 && spare_group != NULL
@@ -347,8 +382,12 @@ int Monitor(mddev_dev_t devlist,
 						close(fd2);
 					}
 			}
-					    
-		sleep(period);
+		if (!new_found) {
+			if (oneshot)
+				break;
+			else
+				sleep(period);
+		}
 	}
 	return 0;
 }
@@ -374,7 +413,9 @@ static void alert(char *event, char *dev, char *disc, char *mailaddr, char *cmd)
 			exit(2);
 		}
 	}
-	if (mailaddr && strncmp(event, "Fail", 4)==0) {
+	if (mailaddr && 
+	    (strncmp(event, "Fail", 4)==0 || 
+	     strncmp(event, "Degrade", 7)==0)) {
 		FILE *mp = popen(Sendmail, "w");
 		if (mp) {
 			char hname[256];
@@ -390,7 +431,7 @@ static void alert(char *event, char *dev, char *disc, char *mailaddr, char *cmd)
 			fprintf(mp, "A %s event had been detected on md device %s.\n\n", event, dev);
 
 			if (disc)
-				fprintf(mp, "It could be related to componenet device %s.\n\n", disc);
+				fprintf(mp, "It could be related to component device %s.\n\n", disc);
 
 			fprintf(mp, "Faithfully yours, etc.\n");
 			fclose(mp);
